"""XML Parser."""

from __future__ import annotations

from datetime import datetime
import logging

import dateutil
import xmltodict

from georss_client.consts import (
    XML_TAG_CHANNEL,
    XML_TAG_DC_DATE,
    XML_TAG_FEED,
    XML_TAG_GEO_LAT,
    XML_TAG_GEO_LONG,
    XML_TAG_GEORSS_POINT,
    XML_TAG_GEORSS_POLYGON,
    XML_TAG_GML_POS,
    XML_TAG_GML_POS_LIST,
    XML_TAG_HEIGHT,
    XML_TAG_LAST_BUILD_DATE,
    XML_TAG_PUB_DATE,
    XML_TAG_PUBLISHED,
    XML_TAG_RSS,
    XML_TAG_TTL,
    XML_TAG_UPDATED,
    XML_TAG_WIDTH,
)
from georss_client.xml_parser.feed import Feed

_LOGGER = logging.getLogger(__name__)

DEFAULT_NAMESPACES = {
    "http://www.w3.org/2005/Atom": None,
    "https://www.w3.org/2005/Atom": None,
    "http://purl.org/dc/elements/1.1/": "dc",
    "https://purl.org/dc/elements/1.1/": "dc",
    "http://www.georss.org/georss": "georss",
    "https://www.georss.org/georss": "georss",
    "http://www.ogc.org/standard/georss/": "georss",
    "https://www.ogc.org/standard/georss/": "georss",
    "http://www.w3.org/2003/01/geo/wgs84_pos#": "geo",
    "https://www.w3.org/2003/01/geo/wgs84_pos#": "geo",
    "http://www.w3.org/2003/01/geo/": "geo",
    "https://www.w3.org/2003/01/geo/": "geo",
    "http://www.opengis.net/gml": "gml",
    "https://www.opengis.net/gml": "gml",
    "http://www.gdacs.org/": "gdacs",
    "https://www.gdacs.org/": "gdacs",
}
KEYS_DATE = [
    XML_TAG_DC_DATE,
    XML_TAG_LAST_BUILD_DATE,
    XML_TAG_PUB_DATE,
    XML_TAG_PUBLISHED,
    XML_TAG_UPDATED,
]
KEYS_FLOAT = [XML_TAG_GEO_LAT, XML_TAG_GEO_LONG]
KEYS_FLOAT_LIST = [
    XML_TAG_GEORSS_POLYGON,
    XML_TAG_GML_POS_LIST,
    XML_TAG_GML_POS,
    XML_TAG_GEORSS_POINT,
]
KEYS_INT = [XML_TAG_HEIGHT, XML_TAG_TTL, XML_TAG_WIDTH]


class XmlParser:
    """Built-in XML parser."""

    def __init__(self, additional_namespaces: dict | None = None):
        """Initialise the XML parser."""
        self._namespaces = DEFAULT_NAMESPACES
        if additional_namespaces:
            self._namespaces.update(additional_namespaces)

    @staticmethod
    def postprocessor(
        path: list[str], key: str, value: str
    ) -> tuple[str, str | float | int | datetime | tuple]:
        """Conduct type conversion for selected keys."""
        try:
            if key in KEYS_DATE and value:
                # Check if value is a dict -> need to extract #text attribute
                if isinstance(value, dict):
                    value = value["#text"]
                return key, dateutil.parser.parse(value)
            if key in KEYS_FLOAT and value:
                return key, float(value)
            if key in KEYS_FLOAT_LIST and value:
                # Check if value is a dict -> need to extract #text attribute
                if isinstance(value, dict):
                    value = value["#text"]
                # Turn white-space separated list of numbers into
                # list of floats.
                coordinate_values = value.split()
                point_coordinates: list[float] = [
                    float(coordinate_values[i]) for i in range(len(coordinate_values))
                ]
                return key, tuple(point_coordinates)
            if key in KEYS_INT and value:
                return key, int(value)
        except (ValueError, TypeError) as error:
            _LOGGER.warning("Unable to process (%s/%s): %s", key, value, error)
        return key, value

    def parse(self, xml: str) -> Feed | None:
        """Parse the provided xml."""
        if xml:
            parsed_dict = xmltodict.parse(
                xml,
                process_namespaces=True,
                namespaces=self._namespaces,
                postprocessor=XmlParser.postprocessor,
            )
            if XML_TAG_RSS in parsed_dict:
                rss = parsed_dict.get(XML_TAG_RSS)
                if XML_TAG_CHANNEL in rss:
                    return Feed(rss.get(XML_TAG_CHANNEL))
            if XML_TAG_FEED in parsed_dict:
                return Feed(parsed_dict.get(XML_TAG_FEED))
        return None



"""GeoRSS feed or feed item."""

from __future__ import annotations

import datetime

from georss_client.consts import (
    XML_ATTR_TERM,
    XML_TAG_AUTHOR,
    XML_TAG_CATEGORY,
    XML_TAG_CONTRIBUTOR,
    XML_TAG_DC_DATE,
    XML_TAG_LAST_BUILD_DATE,
    XML_TAG_MANAGING_EDITOR,
    XML_TAG_NAME,
    XML_TAG_PUB_DATE,
    XML_TAG_PUBLISHED,
    XML_TAG_UPDATED,
)
from georss_client.xml_parser.feed_dict_source import FeedDictSource


class FeedOrFeedItem(FeedDictSource):
    """Represents the common base of feed and its items."""

    @property
    def category(self) -> list | None:
        """Return the categories of this feed item."""
        category = self._attribute([XML_TAG_CATEGORY])
        if category:
            if isinstance(category, (str, dict)):
                # If it's a string or a dict, wrap in list.
                category = [category]
            result = []
            for item in category:
                if XML_ATTR_TERM in item:
                    # <category term="Category 1"/>
                    item = item.get(XML_ATTR_TERM)
                result.append(item)
            return result
        return None

    @property
    def published_date(self) -> datetime.datetime | None:
        """Return the published date of this feed or feed item."""
        parsed_date = self._attribute(
            [XML_TAG_PUB_DATE, XML_TAG_PUBLISHED, XML_TAG_DC_DATE]
        )
        return parsed_date if isinstance(parsed_date, datetime.datetime) else None

    @property
    def pub_date(self) -> datetime.datetime | None:
        """Return the published date of this feed or feed item."""
        return self.published_date

    @property
    def updated_date(self) -> datetime.datetime | None:
        """Return the updated date of this feed or feed item."""
        parsed_date = self._attribute([XML_TAG_LAST_BUILD_DATE, XML_TAG_UPDATED])
        return parsed_date if isinstance(parsed_date, datetime.datetime) else None

    @property
    def last_build_date(self) -> datetime.datetime | None:
        """Return the last build date of this feed."""
        return self.updated_date

    @property
    def author(self) -> str | None:
        """Return the author of this feed."""
        # <managingEditor>jrc-ems@ec.europa.eu</managingEditor>
        managing_editor = self._attribute([XML_TAG_MANAGING_EDITOR])
        if managing_editor:
            return managing_editor
        # <author>
        #   <name>Istituto Nazionale di Geofisica e Vulcanologia</name>
        #   <uri>http://www.ingv.it</uri>
        # </author>
        author = self._attribute([XML_TAG_AUTHOR, XML_TAG_CONTRIBUTOR])
        if author:
            return author.get(XML_TAG_NAME, None)
        return None

    @property
    def contributor(self) -> str | None:
        """Return the contributor of this feed."""
        return self.author

    @property
    def managing_editor(self) -> str | None:
        """Return the managing editor of this feed."""
        return self.author



"""Tests for XML parser."""

import datetime
from pyexpat import ExpatError

import pytest

from georss_client.xml_parser import XmlParser
from georss_client.xml_parser.geometry import Point, Polygon
from tests.utils import load_fixture


def test_simple_1():
    """Test parsing various actual XML files."""
    xml_parser = XmlParser()
    xml = load_fixture("xml_parser_simple_1.xml")
    feed = xml_parser.parse(xml)
    assert feed is not None
    assert feed.entries is not None
    assert len(feed.entries) == 1


def test_simple_2():
    """Test parsing various actual XML files."""
    xml_parser = XmlParser()
    xml = load_fixture("xml_parser_simple_2.xml")
    feed = xml_parser.parse(xml)
    assert feed is not None
    assert feed.entries is not None
    assert len(feed.entries) == 1


def test_simple_3():
    """Test parsing various actual XML files."""
    xml_parser = XmlParser()
    xml = load_fixture("xml_parser_simple_3.xml")
    feed = xml_parser.parse(xml)
    assert feed is None


def test_complex_1():
    """Test parsing various actual XML files."""
    xml_parser = XmlParser()
    xml = load_fixture("xml_parser_complex_1.xml")
    feed = xml_parser.parse(xml)
    assert feed is not None

    assert feed.title == "Feed Title 1"
    assert feed.subtitle == "Feed Subtitle 1"
    assert feed.description == "Feed Description 1"
    assert feed.summary == "Feed Description 1"
    assert feed.content == "Feed Description 1"
    assert feed.link == "Feed Link 1"
    assert feed.published_date == datetime.datetime(
        2018, 12, 9, 8, 30, tzinfo=datetime.timezone.utc
    )
    assert feed.pub_date == datetime.datetime(
        2018, 12, 9, 8, 30, tzinfo=datetime.timezone.utc
    )
    assert feed.updated_date == datetime.datetime(
        2018, 12, 9, 8, 45, tzinfo=datetime.timezone.utc
    )
    assert feed.last_build_date == datetime.datetime(
        2018, 12, 9, 8, 45, tzinfo=datetime.timezone.utc
    )
    assert feed.copyright == "Feed Copyright 1"
    assert feed.rights == "Feed Copyright 1"
    assert feed.generator == "Feed Generator 1"
    assert feed.language == "Feed Language 1"
    assert feed.docs == "http://docs.url/documentation.html"
    assert feed.ttl == 42
    assert feed.author == "Feed Author 1"
    assert feed.contributor == "Feed Author 1"
    assert feed.managing_editor == "Feed Author 1"
    assert feed.category == ["Feed Category 1"]
    assert feed.image is not None
    assert feed.image.title == "Image Title 1"
    assert feed.image.url == "http://image.url/image.png"
    assert feed.image.link == "http://feed.link/feed.rss"
    assert feed.image.description == "Image Description 1"
    assert feed.image.width == 123
    assert feed.image.height == 234
    assert feed.get_additional_attribute("random") == "Feed Random 1"
    assert repr(feed) == "<Feed(Feed Link 1)>"

    assert feed.entries is not None
    assert len(feed.entries) == 6

    feed_entry = feed.entries[0]
    assert feed_entry.title == "Title 1"
    assert feed_entry.description == "Description 1"
    assert feed_entry.link == "Link 1"
    assert feed_entry.published_date == datetime.datetime(
        2018, 12, 9, 7, 30, tzinfo=datetime.timezone.utc
    )
    assert feed_entry.updated_date == datetime.datetime(
        2018, 12, 9, 7, 45, tzinfo=datetime.timezone.utc
    )
    assert feed_entry.guid == "GUID 1"
    assert feed_entry.id == "GUID 1"
    assert feed_entry.source == "Source 1"
    assert feed_entry.category == ["Category 1"]
    assert isinstance(feed_entry.geometry, Point)
    assert feed_entry.geometry.latitude == -37.4567
    assert feed_entry.geometry.longitude == 149.3456
    assert feed_entry.get_additional_attribute("random") == "Random 1"
    assert repr(feed_entry) == "<FeedItem(GUID 1)>"

    feed_entry = feed.entries[1]
    assert feed_entry.title == "Title 2"
    assert feed_entry.description == "Description 2"
    assert feed_entry.link == "Link 2"
    assert feed_entry.published_date == datetime.datetime(
        2018, 12, 9, 7, 35, tzinfo=datetime.timezone.utc
    )
    assert feed_entry.updated_date == datetime.datetime(
        2018, 12, 9, 7, 50, tzinfo=datetime.timezone.utc
    )
    assert feed_entry.guid == "GUID 2"
    assert feed_entry.category == ["Category 2"]
    assert isinstance(feed_entry.geometry, Point)
    assert feed_entry.geometry.latitude == -37.5678
    assert feed_entry.geometry.longitude == 149.4567

    feed_entry = feed.entries[2]
    assert feed_entry.title == "Title 3"
    assert feed_entry.description == "Description 3"
    assert feed_entry.published_date == datetime.datetime(
        2018, 12, 9, 7, 40, tzinfo=datetime.timezone.utc
    )
    assert feed_entry.updated_date == datetime.datetime(
        2018, 12, 9, 7, 55, tzinfo=datetime.timezone.utc
    )
    assert feed_entry.guid == "GUID 3"
    assert feed_entry.category == ["Category 3A", "Category 3B", "Category 3C"]
    assert isinstance(feed_entry.geometry, Point)
    assert feed_entry.geometry.latitude == -37.6789
    assert feed_entry.geometry.longitude == 149.5678

    feed_entry = feed.entries[3]
    assert feed_entry.title == "Title 4"
    assert feed_entry.description == "Description 4"
    assert feed_entry.author == "Author 4"
    assert feed_entry.contributor == "Author 4"
    assert feed_entry.category == ["Category 4A", "Category 4B"]
    assert feed_entry.published_date == datetime.datetime(
        2018,
        9,
        30,
        21,
        36,
        48,
        tzinfo=datetime.timezone(datetime.timedelta(hours=10), "AEST"),
    )
    assert isinstance(feed_entry.geometry, Point)
    assert feed_entry.geometry.latitude == -37.789
    assert feed_entry.geometry.longitude == 149.6789

    feed_entry = feed.entries[4]
    assert feed_entry.title == "Title 5"
    assert feed_entry.description == "Description 5"
    assert feed_entry.published_date == datetime.datetime(
        2018,
        9,
        20,
        18,
        1,
        55,
        tzinfo=datetime.timezone(datetime.timedelta(hours=2), "CEST"),
    )
    assert isinstance(feed_entry.geometry, Polygon)
    assert feed_entry.geometry.centroid.latitude == -30.32
    assert feed_entry.geometry.centroid.longitude == 150.32

    feed_entry = feed.entries[5]
    assert feed_entry.title == "Title 6"
    assert feed_entry.description == "Description 6"
    assert feed_entry.published_date == datetime.datetime(
        2018,
        10,
        7,
        19,
        52,
        tzinfo=datetime.timezone(datetime.timedelta(hours=-2)),
    )
    assert isinstance(feed_entry.geometry, Polygon)
    assert feed_entry.geometry.centroid.latitude == -30.32
    assert feed_entry.geometry.centroid.longitude == 150.32


def test_complex_2():
    """Test parsing various actual XML files."""
    xml_parser = XmlParser()
    xml = load_fixture("xml_parser_complex_2.xml")
    feed = xml_parser.parse(xml)
    assert feed is not None

    assert feed.title == "Feed Title 1"
    assert feed.subtitle == "Feed Subtitle 1"
    assert feed.ttl == "INVALID"
    assert feed.author == "Author 1"
    assert feed.last_build_date == datetime.datetime(
        2018, 12, 9, 9, 0, tzinfo=datetime.timezone.utc
    )
    assert feed.updated_date == datetime.datetime(
        2018, 12, 9, 9, 0, tzinfo=datetime.timezone.utc
    )
    assert feed.copyright == "Feed Rights 1"
    assert feed.rights == "Feed Rights 1"
    assert feed.generator == "Feed Generator 1"
    assert feed.image is not None
    assert feed.image.title == "Image Title 1"
    assert feed.image.url == "http://image.url/image.png"
    assert feed.image.link == "http://feed.link/feed.rss"
    assert feed.image.description is None
    assert feed.image.width is None
    assert feed.image.height is None
    assert feed.docs is None

    assert feed.entries is not None
    assert len(feed.entries) == 1

    feed_entry = feed.entries[0]
    assert feed_entry.title == "Title 6"
    assert feed_entry.published_date is None


def test_complex_3():
    """Test parsing various actual XML files."""
    xml_parser = XmlParser()
    xml = load_fixture("xml_parser_complex_3.xml")
    feed = xml_parser.parse(xml)
    assert feed is not None

    assert feed.title is None
    assert feed.subtitle is None
    assert feed.description is None
    assert feed.language is None
    assert feed.published_date is None
    assert feed.last_build_date is None
    assert feed.ttl is None
    assert feed.rights == "Feed Rights 1"
    assert feed.image is None

    assert feed.entries is not None
    assert len(feed.entries) == 2

    feed_entry = feed.entries[0]
    assert feed_entry.title is None
    assert feed_entry.published_date is None
    assert feed_entry.geometry is None

    feed_entry = feed.entries[1]
    assert feed_entry.title is None
    assert feed_entry.geometry is None


def test_geometries_1():
    """Test parsing various geometries in entries."""
    xml_parser = XmlParser()
    xml = load_fixture("xml_parser_geometries_1.xml")
    feed = xml_parser.parse(xml)
    assert feed is not None

    assert feed.title == "Feed Title 1"
    assert feed.entries is not None
    assert len(feed.entries) == 6

    feed_entry = feed.entries[0]
    assert feed_entry.title == "Title 1"
    assert feed_entry.geometries is not None
    assert len(feed_entry.geometries) == 3

    feed_entry = feed.entries[1]
    assert feed_entry.title == "Title 2"
    assert feed_entry.geometries is not None
    assert len(feed_entry.geometries) == 3

    feed_entry = feed.entries[2]
    assert feed_entry.title == "Title 3"
    assert feed_entry.geometries is not None
    assert len(feed_entry.geometries) == 2

    feed_entry = feed.entries[3]
    assert feed_entry.title == "Title 4"
    assert feed_entry.geometries is not None
    assert len(feed_entry.geometries) == 2

    feed_entry = feed.entries[4]
    assert feed_entry.title == "Title 5"
    assert feed_entry.geometries is not None
    assert len(feed_entry.geometries) == 3

    feed_entry = feed.entries[5]
    assert feed_entry.title == "Title 6"
    assert feed_entry.geometries is not None
    assert len(feed_entry.geometries) == 2


def test_geometries_2():
    """Test parsing various geometries in entries."""
    xml_parser = XmlParser()
    xml = load_fixture("xml_parser_geometries_2.xml")
    feed = xml_parser.parse(xml)
    assert feed is not None

    assert feed.title == "Feed Title 1"
    assert feed.entries is not None
    assert len(feed.entries) == 6

    feed_entry = feed.entries[0]
    assert feed_entry.title == "Title 1"
    assert feed_entry.geometries is not None
    assert len(feed_entry.geometries) == 3

    feed_entry = feed.entries[1]
    assert feed_entry.title == "Title 2"
    assert feed_entry.geometries is not None
    assert len(feed_entry.geometries) == 3

    feed_entry = feed.entries[2]
    assert feed_entry.title == "Title 3"
    assert feed_entry.geometries is not None
    assert len(feed_entry.geometries) == 2

    feed_entry = feed.entries[3]
    assert feed_entry.title == "Title 4"
    assert feed_entry.geometries is not None
    assert len(feed_entry.geometries) == 2

    feed_entry = feed.entries[4]
    assert feed_entry.title == "Title 5"
    assert feed_entry.geometries is not None
    assert len(feed_entry.geometries) == 3

    feed_entry = feed.entries[5]
    assert feed_entry.title == "Title 6"
    assert feed_entry.geometries is not None
    assert len(feed_entry.geometries) == 2


def test_byte_order_mark():
    """Test parsing an XML file with byte order mark."""
    xml_parser = XmlParser()
    # Create XML starting with byte order mark.
    xml = (
        "\xef\xbb\xbf<?xml version='1.0' encoding='utf-8'?>"
        "<rss version='2.0'><channel><item><title>Title 1</title>"
        "</item></channel></rss>"
    )
    # This will raise an error because the parser can't handle
    with pytest.raises(ExpatError):
        xml_parser.parse(xml)



"""Tests for georss distance helper."""

from unittest.mock import MagicMock

import pytest

from georss_client.geo_rss_distance_helper import GeoRssDistanceHelper
from georss_client.xml_parser.geometry import Point, Polygon


def test_extract_coordinates_from_point():
    """Test extracting coordinates from point."""
    mock_point = Point(-30.0, 151.0)
    latitude, longitude = GeoRssDistanceHelper.extract_coordinates(mock_point)
    assert latitude == -30.0
    assert longitude == 151.0


def test_extract_coordinates_from_polygon():
    """Test extracting coordinates from polygon."""
    mock_polygon = Polygon(
        [
            Point(-30.0, 151.0),
            Point(-30.0, 151.5),
            Point(-30.5, 151.5),
            Point(-30.5, 151.0),
            Point(-30.0, 151.0),
        ]
    )
    latitude, longitude = GeoRssDistanceHelper.extract_coordinates(mock_polygon)
    assert latitude == pytest.approx(-30.2, 0.1)
    assert longitude == pytest.approx(151.2, 0.1)


def test_extract_coordinates_from_unsupported_geometry():
    """Test extracting coordinates from unsupported geometry."""
    mock_unsupported_geometry = MagicMock()
    latitude, longitude = GeoRssDistanceHelper.extract_coordinates(
        mock_unsupported_geometry
    )
    assert latitude is None
    assert longitude is None


def test_distance_to_point():
    """Test calculating distance to point."""
    home_coordinates = [-31.0, 150.0]
    mock_point = Point(-30.0, 151.0)
    distance = GeoRssDistanceHelper.distance_to_geometry(home_coordinates, mock_point)
    assert distance == pytest.approx(146.8, 0.1)


def test_distance_to_polygon():
    """Test calculating distance to point."""
    home_coordinates = [-31.0, 150.0]
    mock_polygon = Polygon(
        [
            Point(-30.0, 151.0),
            Point(-30.0, 151.5),
            Point(-30.5, 151.5),
            Point(-30.5, 151.0),
            Point(-30.0, 151.0),
        ]
    )
    distance = GeoRssDistanceHelper.distance_to_geometry(home_coordinates, mock_polygon)
    assert distance == pytest.approx(110.6, 0.1)


def test_distance_to_unsupported_geometry():
    """Test calculating distance to unsupported geometry."""
    home_coordinates = [-31.0, 150.0]
    mock_unsupported_geometry = MagicMock()
    distance = GeoRssDistanceHelper.distance_to_geometry(
        home_coordinates, mock_unsupported_geometry
    )
    assert distance == float("inf")



"""Geometry models."""

from __future__ import annotations


class Geometry:
    """Represents a geometry."""


class Point(Geometry):
    """Represents a point."""

    def __init__(self, latitude, longitude):
        """Initialise point."""
        self._latitude = latitude
        self._longitude = longitude

    def __repr__(self):
        """Return string representation of this point."""
        return f"<{self.__class__.__name__}(latitude={self.latitude}, longitude={self.longitude})>"

    def __hash__(self) -> int:
        """Return unique hash of this geometry."""
        return hash((self.latitude, self.longitude))

    def __eq__(self, other: object) -> bool:
        """Return if this object is equal to other object."""
        return (
            self.__class__ == other.__class__
            and self.latitude == other.latitude
            and self.longitude == other.longitude
        )

    @property
    def latitude(self) -> float | None:
        """Return the latitude of this point."""
        return self._latitude

    @property
    def longitude(self) -> float | None:
        """Return the longitude of this point."""
        return self._longitude


class Polygon(Geometry):
    """Represents a polygon."""

    def __init__(self, points: list[Point]):
        """Initialise polygon."""
        self._points: list[Point] = points

    def __repr__(self):
        """Return string representation of this polygon."""
        return f"<{self.__class__.__name__}(centroid={self.centroid})>"

    def __hash__(self) -> int:
        """Return unique hash of this geometry."""
        return hash(self.points)

    def __eq__(self, other: object) -> bool:
        """Return if this object is equal to other object."""
        return self.__class__ == other.__class__ and self.points == other.points

    @property
    def points(self) -> list[Point] | None:
        """Return the points of this polygon."""
        return self._points

    @property
    def centroid(self) -> Point:
        """Find the polygon's centroid as a best approximation."""
        longitudes_list: list[float] = [point.longitude for point in self.points]
        latitudes_list: list[float] = [point.latitude for point in self.points]
        number_of_points: int = len(self.points)
        longitude: float = sum(longitudes_list) / number_of_points
        latitude: float = sum(latitudes_list) / number_of_points
        return Point(latitude, longitude)



"""Test for the Feed Manager."""

import datetime
from unittest import mock

import pytest

from georss_client.feed_manager import FeedManagerBase
from tests import MockGeoRssFeed
from tests.utils import load_fixture

HOME_COORDINATES_1 = (-31.0, 151.0)
HOME_COORDINATES_2 = (-37.0, 150.0)


@mock.patch("requests.Request")
@mock.patch("requests.Session")
def test_feed_manager(mock_session, mock_request):
    """Test the feed manager."""
    mock_session.return_value.__enter__.return_value.send.return_value.ok = True
    mock_session.return_value.__enter__.return_value.send.return_value.text = (
        load_fixture("generic_feed_1.xml")
    )

    feed = MockGeoRssFeed(HOME_COORDINATES_1, None)

    # This will just record calls and keep track of external ids.
    generated_entity_external_ids = []
    updated_entity_external_ids = []
    removed_entity_external_ids = []

    def _generate_entity(external_id):
        """Generate new entity."""
        generated_entity_external_ids.append(external_id)

    def _update_entity(external_id):
        """Update entity."""
        updated_entity_external_ids.append(external_id)

    def _remove_entity(external_id):
        """Remove entity."""
        removed_entity_external_ids.append(external_id)

    feed_manager = FeedManagerBase(
        feed, _generate_entity, _update_entity, _remove_entity
    )
    assert (
        repr(feed_manager) == "<FeedManagerBase("
        "feed=<MockGeoRssFeed(home="
        "(-31.0, 151.0), url=None, "
        "radius=None, categories=None)>)>"
    )
    feed_manager.update()
    entries = feed_manager.feed_entries
    assert entries is not None
    assert len(entries) == 5
    assert feed_manager.last_update is not None
    assert feed_manager.last_timestamp == datetime.datetime(2018, 9, 23, 9, 10)
    assert len(generated_entity_external_ids) == 5
    assert len(updated_entity_external_ids) == 0
    assert len(removed_entity_external_ids) == 0

    feed_entry = entries.get("1234")
    assert feed_entry.title == "Title 1"
    assert feed_entry.external_id == "1234"
    assert feed_entry.coordinates == (-37.2345, 149.1234)
    assert feed_entry.distance_to_home == pytest.approx(714.4, 0.1)
    assert repr(feed_entry) == "<FeedEntry(id=1234)>"

    feed_entry = entries.get("2345")
    assert feed_entry.title == "Title 2"
    assert feed_entry.external_id == "2345"

    feed_entry = entries.get("Title 3")
    assert feed_entry.title == "Title 3"
    assert feed_entry.external_id == "Title 3"

    external_id = hash((-37.8901, 149.7890))
    feed_entry = entries.get(external_id)
    assert feed_entry.title is None
    assert feed_entry.external_id == external_id

    feed_entry = entries.get("5678")
    assert feed_entry.title == "Title 5"
    assert feed_entry.external_id == "5678"

    # Simulate an update with several changes.
    generated_entity_external_ids.clear()
    updated_entity_external_ids.clear()
    removed_entity_external_ids.clear()

    mock_session.return_value.__enter__.return_value.send.return_value.text = (
        load_fixture("generic_feed_4.xml")
    )

    feed_manager.update()
    entries = feed_manager.feed_entries
    assert entries is not None
    assert len(entries) == 3
    assert len(generated_entity_external_ids) == 1
    assert len(updated_entity_external_ids) == 2
    assert len(removed_entity_external_ids) == 3

    feed_entry = entries.get("1234")
    assert feed_entry.title == "Title 1 UPDATED"

    feed_entry = entries.get("2345")
    assert feed_entry.title == "Title 2"

    feed_entry = entries.get("6789")
    assert feed_entry.title == "Title 6"

    # Simulate an update with no data.
    generated_entity_external_ids.clear()
    updated_entity_external_ids.clear()
    removed_entity_external_ids.clear()

    mock_session.return_value.__enter__.return_value.send.return_value.ok = False

    feed_manager.update()
    entries = feed_manager.feed_entries

    assert len(entries) == 0
    assert len(generated_entity_external_ids) == 0
    assert len(updated_entity_external_ids) == 0
    assert len(removed_entity_external_ids) == 3


@mock.patch("requests.Request")
@mock.patch("requests.Session")
def test_feed_manager_no_timestamp(mock_session, mock_request):
    """Test updating feed is ok."""
    mock_session.return_value.__enter__.return_value.send.return_value.ok = True
    mock_session.return_value.__enter__.return_value.send.return_value.text = (
        load_fixture("generic_feed_5.xml")
    )

    feed = MockGeoRssFeed(HOME_COORDINATES_1, None)

    # This will just record calls and keep track of external ids.
    generated_entity_external_ids = []
    updated_entity_external_ids = []
    removed_entity_external_ids = []

    def _generate_entity(external_id):
        """Generate new entity."""
        generated_entity_external_ids.append(external_id)

    def _update_entity(external_id):
        """Update entity."""
        updated_entity_external_ids.append(external_id)

    def _remove_entity(external_id):
        """Remove entity."""
        removed_entity_external_ids.append(external_id)

    feed_manager = FeedManagerBase(
        feed, _generate_entity, _update_entity, _remove_entity
    )
    assert (
        repr(feed_manager) == "<FeedManagerBase("
        "feed=<MockGeoRssFeed(home="
        "(-31.0, 151.0), url=None, "
        "radius=None, categories=None)>)>"
    )
    feed_manager.update()
    entries = feed_manager.feed_entries
    assert entries is not None
    assert len(entries) == 1
    assert feed_manager.last_timestamp is None



"""Base class for the feed manager.

This allows managing feeds and their entries throughout their life-cycle.
"""

from __future__ import annotations

from datetime import datetime
import logging
from typing import Callable

from . import GeoRssFeed
from .consts import UPDATE_OK, UPDATE_OK_NO_DATA

_LOGGER = logging.getLogger(__name__)


class FeedManagerBase:
    """Generic Feed manager."""

    def __init__(
        self,
        feed: GeoRssFeed,
        generate_callback: Callable[[str], None],
        update_callback: Callable[[str], None],
        remove_callback: Callable[[str], None],
    ):
        """Initialise feed manager."""
        self._feed: GeoRssFeed = feed
        self.feed_entries: dict = {}
        self._managed_external_ids = set()
        self._last_update: datetime | None = None
        self._generate_callback: Callable[[str], None] = generate_callback
        self._update_callback: Callable[[str], None] = update_callback
        self._remove_callback: Callable[[str], None] = remove_callback

    def __repr__(self):
        """Return string representation of this feed."""
        return f"<{self.__class__.__name__}(feed={self._feed})>"

    def update(self):
        """Update the feed and then update connected entities."""
        status, feed_entries = self._feed.update()
        if status == UPDATE_OK:
            _LOGGER.debug("Data retrieved %s", feed_entries)
            # Keep a copy of all feed entries for future lookups by entities.
            self.feed_entries = {entry.external_id: entry for entry in feed_entries}
            # Record current time of update.
            self._last_update = datetime.now()
            # For entity management the external ids from the feed are used.
            feed_external_ids = set(self.feed_entries)
            remove_external_ids = self._managed_external_ids.difference(
                feed_external_ids
            )
            self._remove_entities(remove_external_ids)
            update_external_ids = self._managed_external_ids.intersection(
                feed_external_ids
            )
            self._update_entities(update_external_ids)
            create_external_ids = feed_external_ids.difference(
                self._managed_external_ids
            )
            self._generate_new_entities(create_external_ids)
        elif status == UPDATE_OK_NO_DATA:
            _LOGGER.debug("Update successful, but no data received from %s", self._feed)
        else:
            _LOGGER.warning(
                "Update not successful, no data received from %s", self._feed
            )
            # Remove all entities.
            self._remove_entities(self._managed_external_ids.copy())
            # Remove all feed entries and managed external ids.
            self.feed_entries.clear()
            self._managed_external_ids.clear()

    def _generate_new_entities(self, external_ids):
        """Generate new entities for events."""
        for external_id in external_ids:
            self._generate_callback(external_id)
            _LOGGER.debug("New entity added %s", external_id)
            self._managed_external_ids.add(external_id)

    def _update_entities(self, external_ids):
        """Update entities."""
        for external_id in external_ids:
            _LOGGER.debug("Existing entity found %s", external_id)
            self._update_callback(external_id)

    def _remove_entities(self, external_ids):
        """Remove entities."""
        for external_id in external_ids:
            _LOGGER.debug("Entity not current anymore %s", external_id)
            self._managed_external_ids.remove(external_id)
            self._remove_callback(external_id)

    @property
    def last_timestamp(self) -> datetime | None:
        """Return the last timestamp extracted from this feed."""
        return self._feed.last_timestamp

    @property
    def last_update(self) -> datetime | None:
        """Return the last successful update of this feed."""
        return self._last_update



"""GeoRSS Distance Helper."""

from __future__ import annotations

import logging

from haversine import haversine

from .xml_parser.geometry import Geometry, Point, Polygon

_LOGGER = logging.getLogger(__name__)


class GeoRssDistanceHelper:
    """Helper to calculate distances between GeoRSS geometries."""

    @staticmethod
    def extract_coordinates(geometry: Geometry) -> tuple[float, float] | None:
        """Extract the best coordinates from the feature for display."""
        latitude = longitude = None
        if isinstance(geometry, Point):
            # Just extract latitude and longitude directly.
            latitude, longitude = geometry.latitude, geometry.longitude
        elif isinstance(geometry, Polygon):
            centroid = geometry.centroid
            latitude, longitude = centroid.latitude, centroid.longitude
            _LOGGER.debug("Centroid of %s is %s", geometry, (latitude, longitude))
        else:
            _LOGGER.debug("Not implemented: %s", type(geometry))
        return latitude, longitude

    @staticmethod
    def distance_to_geometry(
        home_coordinates: tuple[float, float], geometry: Geometry
    ) -> float:
        """Calculate the distance between home coordinates and geometry."""
        distance: float = float("inf")
        if isinstance(geometry, Point):
            distance = GeoRssDistanceHelper._distance_to_point(
                home_coordinates, geometry
            )
        elif isinstance(geometry, Polygon):
            distance = GeoRssDistanceHelper._distance_to_polygon(
                home_coordinates, geometry
            )
        else:
            _LOGGER.debug("Not implemented: %s", type(geometry))
        return distance

    @staticmethod
    def _distance_to_point(
        home_coordinates: tuple[float, float], point: Point
    ) -> float:
        """Calculate the distance between home coordinates and the point."""
        # Swap coordinates to match: (latitude, longitude).
        return GeoRssDistanceHelper._distance_to_coordinates(
            home_coordinates, (point.latitude, point.longitude)
        )

    @staticmethod
    def _distance_to_polygon(
        home_coordinates: tuple[float, float], polygon: Polygon
    ) -> float:
        """Calculate the distance between home coordinates and the polygon."""
        distance = float("inf")
        # Calculate distance from polygon by calculating the distance
        # to each point of the polygon but not to each edge of the
        # polygon; should be good enough
        for point in polygon.points:
            distance = min(
                distance,
                GeoRssDistanceHelper._distance_to_coordinates(
                    home_coordinates, (point.latitude, point.longitude)
                ),
            )
        return distance

    @staticmethod
    def _distance_to_coordinates(
        home_coordinates: tuple[float, float], coordinates: tuple[float, float]
    ) -> float:
        """Calculate the distance between home coordinates and the coordinates."""
        # Expecting coordinates in format: (latitude, longitude).
        return haversine(coordinates, home_coordinates)



"""Feed Entry."""

from __future__ import annotations

from datetime import datetime
import re

from .consts import CUSTOM_ATTRIBUTE
from .geo_rss_distance_helper import GeoRssDistanceHelper
from .xml_parser.feed_item import FeedItem
from .xml_parser.geometry import Geometry


class FeedEntry:
    """Feed entry base class."""

    def __init__(self, home_coordinates: tuple[float, float], rss_entry: FeedItem):
        """Initialise this feed entry."""
        self._home_coordinates: tuple[float, float] = home_coordinates
        self._rss_entry: FeedItem = rss_entry

    def __repr__(self):
        """Return string representation of this entry."""
        return f"<{self.__class__.__name__}(id={self.external_id})>"

    @property
    def geometry(self) -> Geometry | None:
        """Return all geometry details of this entry."""
        if self._rss_entry:
            return self._rss_entry.geometry
        return None

    @property
    def coordinates(self) -> tuple[float, float] | None:
        """Return the best coordinates (latitude, longitude) of this entry."""
        if self.geometry:
            return GeoRssDistanceHelper.extract_coordinates(self.geometry)
        return None

    @property
    def external_id(self) -> str | None:
        """Return the external id of this entry."""
        if self._rss_entry:
            external_id = self._rss_entry.guid
            if not external_id:
                external_id = self.title
            if not external_id:
                # Use geometry as ID as a fallback.
                external_id = hash(self.coordinates)
            return external_id
        return None

    def _search_in_external_id(self, regexp) -> str | None:
        """Find a sub-string in the entry's external id."""
        if self.external_id:
            match = re.search(regexp, self.external_id)
            if match:
                return match.group(CUSTOM_ATTRIBUTE)
        return None

    @property
    def title(self) -> str | None:
        """Return the title of this entry."""
        if self._rss_entry:
            return self._rss_entry.title
        return None

    def _search_in_title(self, regexp):
        """Find a sub-string in the entry's title."""
        if self.title:
            match = re.search(regexp, self.title)
            if match:
                return match.group(CUSTOM_ATTRIBUTE)
        return None

    @property
    def category(self) -> str | None:
        """Return the category of this entry."""
        if (
            self._rss_entry
            and self._rss_entry.category
            and isinstance(self._rss_entry.category, list)
        ):
            # To keep this simple, just return the first category.
            return self._rss_entry.category[0]
        return None

    @property
    def attribution(self) -> str | None:
        """Return the attribution of this entry."""
        return None

    @property
    def distance_to_home(self) -> float:
        """Return the distance in km of this entry to the home coordinates."""
        return GeoRssDistanceHelper.distance_to_geometry(
            self._home_coordinates, self.geometry
        )

    @property
    def description(self) -> str | None:
        """Return the description of this entry."""
        if self._rss_entry and self._rss_entry.description:
            return self._rss_entry.description
        return None

    @property
    def published(self) -> datetime | None:
        """Return the published date of this entry."""
        if self._rss_entry:
            return self._rss_entry.published_date
        return None

    @property
    def updated(self) -> datetime | None:
        """Return the updated date of this entry."""
        if self._rss_entry:
            return self._rss_entry.updated_date
        return None

    def _search_in_description(self, regexp):
        """Find a sub-string in the entry's description."""
        if self.description:
            match = re.search(regexp, self.description)
            if match:
                return match.group(CUSTOM_ATTRIBUTE)
        return None



"""GeoRSS feed item."""

from __future__ import annotations

from georss_client.consts import (
    XML_TAG_GEO_LAT,
    XML_TAG_GEO_LONG,
    XML_TAG_GEO_POINT,
    XML_TAG_GEORSS_POINT,
    XML_TAG_GEORSS_POLYGON,
    XML_TAG_GEORSS_WHERE,
    XML_TAG_GML_EXTERIOR,
    XML_TAG_GML_LINEAR_RING,
    XML_TAG_GML_POINT,
    XML_TAG_GML_POLYGON,
    XML_TAG_GML_POS,
    XML_TAG_GML_POS_LIST,
    XML_TAG_GUID,
    XML_TAG_ID,
    XML_TAG_SOURCE,
)
from georss_client.xml_parser.feed_or_feed_item import FeedOrFeedItem
from georss_client.xml_parser.geometry import Geometry, Point, Polygon


class FeedItem(FeedOrFeedItem):
    """Represents a feed item."""

    def __repr__(self):
        """Return string representation of this feed item."""
        return f"<{self.__class__.__name__}({self.guid})>"

    @property
    def guid(self) -> str | None:
        """Return the guid of this feed item."""
        return self._attribute_with_text([XML_TAG_GUID, XML_TAG_ID])

    @property
    def id(self) -> str | None:
        """Return the id of this feed item."""
        return self.guid

    @property
    def source(self) -> str | None:
        """Return the source of this feed item."""
        return self._attribute([XML_TAG_SOURCE])

    @property
    def geometries(self) -> list[Geometry] | None:
        """Return all geometries of this feed item."""
        geometries = []
        for entry in [
            self._geometry_georss_point(),
            self._geometry_georss_where(),
            self._geometry_geo_point(),
            self._geometry_geo_long_lat(),
            self._geometry_georss_polygon(),
        ]:
            if entry:
                geometries.extend(entry)
        # Filter out any duplicates.
        unique_geometries = []
        for i in geometries:
            if i not in unique_geometries:
                unique_geometries.append(i)
        return unique_geometries

    def _geometry_georss_point(self) -> list[Point] | None:
        """Check for georss:point tag."""
        # <georss:point>-0.5 119.8</georss:point>
        point = self._attribute([XML_TAG_GEORSS_POINT])
        if point:
            if isinstance(point, tuple):
                return FeedItem._create_georss_point_single(point)
            return FeedItem._create_georss_point_multiple(point)
        return None

    @staticmethod
    def _create_georss_point_single(point: tuple) -> list[Point]:
        """Create single point from provided coordinates."""
        return [Point(point[0], point[1])]

    @staticmethod
    def _create_georss_point_multiple(point: list) -> list[Point]:
        """Create multiple points from provided coordinates."""
        return [Point(entry[0], entry[1]) for entry in point]

    def _geometry_georss_where(self) -> list[Geometry] | None:
        """Check for georss:where tag."""
        where = self._attribute([XML_TAG_GEORSS_WHERE])
        if where:
            # Point:
            # <georss:where>
            #   <gml:Point>
            #     <gml:pos>44.11 -66.23</gml:pos>
            #   </gml:Point>
            # </georss:where>
            pos = self._attribute_in_structure(
                where, [XML_TAG_GML_POINT, XML_TAG_GML_POS]
            )
            if pos:
                return [Point(pos[0], pos[1])]
            # Polygon:
            # <georss:where>
            #   <gml:Polygon>
            #     <gml:exterior>
            #       <gml:LinearRing>
            #         <gml:posList>
            #           -71.106216 42.366661
            #           -71.105576 42.367104
            #           -71.104378 42.367134
            #           -71.103729 42.366249
            #           -71.098793 42.363331
            #           -71.101028 42.362541
            #           -71.106865 42.366123
            #           -71.106216 42.366661
            #         </gml:posList>
            #       </gml:LinearRing>
            #     </gml:exterior>
            #   </gml:Polygon>
            # </georss:where>
            pos_list = self._attribute_in_structure(
                where,
                [
                    XML_TAG_GML_POLYGON,
                    XML_TAG_GML_EXTERIOR,
                    XML_TAG_GML_LINEAR_RING,
                    XML_TAG_GML_POS_LIST,
                ],
            )
            if pos_list:
                return self._create_polygon(pos_list)
        return None

    def _geometry_geo_point(self) -> list[Point] | None:
        """Check for geo:Point tag."""
        # <geo:Point xmlns:geo="http://www.w3.org/2003/01/geo/wgs84_pos#">
        #   <geo:lat>38.3728</geo:lat>
        #   <geo:long>15.7213</geo:long>
        # </geo:Point>
        point = self._attribute([XML_TAG_GEO_POINT])
        if point:
            lat = point.get(XML_TAG_GEO_LAT)
            long = point.get(XML_TAG_GEO_LONG)
            if long and lat:
                return [Point(lat, long)]
        return None

    def _geometry_geo_long_lat(self) -> list[Point] | None:
        """Check for geo:long and geo:lat tags."""
        # <geo:long>119.948006</geo:long>
        # <geo:lat>-23.126413</geo:lat>
        lat = self._attribute([XML_TAG_GEO_LAT])
        long = self._attribute([XML_TAG_GEO_LONG])
        if long and lat:
            return [Point(lat, long)]
        return None

    def _geometry_georss_polygon(self) -> list[Polygon] | None:
        """Check for georss:polygon tag."""
        # <georss:polygon>
        #   -34.937663524 148.597260613
        #   -34.9377026399999 148.597169138
        #   -34.9377002169999 148.59708737
        #   -34.9376945989999 148.59705595
        #   -34.9376863529999 148.596955098
        #   -34.937663524 148.597260613
        # </georss:polygon>
        polygon = self._attribute([XML_TAG_GEORSS_POLYGON])
        if polygon:
            return self._create_polygon(polygon)
        return None

    @staticmethod
    def _create_polygon(polygon_data) -> list[Polygon] | None:
        """Create a polygon from the provided coordinates."""
        if polygon_data:
            # Either tuple or an array of tuples.
            if isinstance(polygon_data, tuple):
                return FeedItem._create_polygon_single(polygon_data)
            return FeedItem._create_polygon_multiple(polygon_data)
        return None

    @staticmethod
    def _create_polygon_single(polygon_data: tuple) -> list[Polygon]:
        """Create polygon from provided tuple of coordinates."""
        if len(polygon_data) % 2 != 0:
            # Not even number of coordinates - chop last entry.
            polygon_data = polygon_data[0 : len(polygon_data) - 1]
        return [
            Polygon(
                [
                    Point(polygon_data[i], polygon_data[i + 1])
                    for i in range(0, len(polygon_data), 2)
                ]
            )
        ]

    @staticmethod
    def _create_polygon_multiple(polygon_data: list) -> list[Polygon]:
        """Create polygon from provided list of coordinates."""
        polygons = []
        for entry in polygon_data:
            polygons.extend(FeedItem._create_polygon(entry))
        return polygons

    @property
    def geometry(self) -> Geometry | None:
        """Return the first geometry of this feed item for backwards compatibility reasons."""
        return self.geometries[0] if self.geometries else None
